import os
import sys
import json
import torchvision
from .data_augmentor import RandomCropVideo, RandomHorizontalFlipVideo, RandomReverseTimeVideo, RandomRotationVideo, IdentityTransform, Scale, ComposeMix
from torch.utils.data import Dataset
from .data_loader_skvideo import VideoFolder


class SomethingSomethingV2(VideoFolder):
    def __init__(self, root, mode='train'):
        upscale_size = int(84 * 1.1)
        assert mode in ['train', 'test', 'validation']
        labels_path = os.path.join(root, 'something-something-v2-labels.json')
        mode_path = os.path.join(root, f'something-something-v2-{mode}.json')
        videos_path = os.path.join(root, '20bn-something-something-v2')

        transform_pre = ComposeMix([
            [RandomRotationVideo(20), "vid"],
            [Scale(upscale_size), "img"],
            [RandomCropVideo(84), "vid"],
            [RandomHorizontalFlipVideo(0), "vid"],
            # [RandomReverseTimeVideo(1), "vid"],
            # [torchvision.transforms.ToTensor(), "img"],
        ])
        # identity transform
        transform_post = ComposeMix([
            [torchvision.transforms.ToTensor(), "img"],
        ])
        super(SomethingSomethingV2, self).__init__(root=videos_path,
                                                   json_file_input=mode_path,
                                                   json_file_labels=labels_path,
                                                   is_val=(mode != 'train'),
                                                   transform_post=transform_post,
                                                   transform_pre=transform_pre,
                                                   clip_size=36,
                                                   nclips=1,
                                                   step_size=1)

    def __getitem__(self, index):
        data = super(SomethingSomethingV2, self).__getitem__(index)
        video, label = data
        return data



import torch
from torch import nn
from torch.nn import functional as F


class _NonLocalBlockND(nn.Module):
    def __init__(self, in_channels, inter_channels=None, dimension=3, sub_sample=True, bn_layer=True):
        """
        :param in_channels:
        :param inter_channels:
        :param dimension:
        :param sub_sample:
        :param bn_layer:
        """

        super(_NonLocalBlockND, self).__init__()

        assert dimension in [1, 2, 3]

        self.dimension = dimension
        self.sub_sample = sub_sample

        self.in_channels = in_channels
        self.inter_channels = inter_channels

        if self.inter_channels is None:
            self.inter_channels = in_channels // 2
            if self.inter_channels == 0:
                self.inter_channels = 1

        if dimension == 3:
            conv_nd = nn.Conv3d
            max_pool_layer = nn.MaxPool3d(kernel_size=(1, 2, 2))
            bn = nn.BatchNorm3d
        elif dimension == 2:
            conv_nd = nn.Conv2d
            max_pool_layer = nn.MaxPool2d(kernel_size=(2, 2))
            bn = nn.BatchNorm2d
        else:
            conv_nd = nn.Conv1d
            max_pool_layer = nn.MaxPool1d(kernel_size=(2))
            bn = nn.BatchNorm1d

        self.g = conv_nd(in_channels=self.in_channels, out_channels=self.inter_channels,
                         kernel_size=1, stride=1, padding=0)

        if bn_layer:
            self.W = nn.Sequential(
                conv_nd(in_channels=self.inter_channels, out_channels=self.in_channels,
                        kernel_size=1, stride=1, padding=0),
                bn(self.in_channels)
            )
            nn.init.constant_(self.W[1].weight, 0)
            nn.init.constant_(self.W[1].bias, 0)
        else:
            self.W = conv_nd(in_channels=self.inter_channels, out_channels=self.in_channels,
                             kernel_size=1, stride=1, padding=0)
            nn.init.constant_(self.W.weight, 0)
            nn.init.constant_(self.W.bias, 0)

        self.theta = conv_nd(in_channels=self.in_channels, out_channels=self.inter_channels,
                             kernel_size=1, stride=1, padding=0)
        self.phi = conv_nd(in_channels=self.in_channels, out_channels=self.inter_channels,
                           kernel_size=1, stride=1, padding=0)

        if sub_sample:
            self.g = nn.Sequential(self.g, max_pool_layer)
            self.phi = nn.Sequential(self.phi, max_pool_layer)

    def forward(self, x, return_nl_map=False):
        """
        :param x: (b, c, t, h, w)
        :param return_nl_map: if True return z, nl_map, else only return z.
        :return:
        """

        batch_size = x.size(0)

        g_x = self.g(x).view(batch_size, self.inter_channels, -1)
        g_x = g_x.permute(0, 2, 1)

        theta_x = self.theta(x).view(batch_size, self.inter_channels, -1)
        theta_x = theta_x.permute(0, 2, 1)
        phi_x = self.phi(x).view(batch_size, self.inter_channels, -1)
        f = torch.matmul(theta_x, phi_x)
        f_div_C = F.softmax(f, dim=-1)

        y = torch.matmul(f_div_C, g_x)
        y = y.permute(0, 2, 1).contiguous()
        y = y.view(batch_size, self.inter_channels, *x.size()[2:])
        W_y = self.W(y)
        z = W_y + x

        if return_nl_map:
            return z, f_div_C
        return z


class NONLocalBlock1D(_NonLocalBlockND):
    def __init__(self, in_channels, inter_channels=None, sub_sample=True, bn_layer=True):
        super(NONLocalBlock1D, self).__init__(in_channels,
                                              inter_channels=inter_channels,
                                              dimension=1, sub_sample=sub_sample,
                                              bn_layer=bn_layer)


class NONLocalBlock2D(_NonLocalBlockND):
    def __init__(self, in_channels, inter_channels=None, sub_sample=True, bn_layer=True):
        super(NONLocalBlock2D, self).__init__(in_channels,
                                              inter_channels=inter_channels,
                                              dimension=2, sub_sample=sub_sample,
                                              bn_layer=bn_layer,)


class NONLocalBlock3D(_NonLocalBlockND):
    def __init__(self, in_channels, inter_channels=None, sub_sample=True, bn_layer=True):
        super(NONLocalBlock3D, self).__init__(in_channels,
                                              inter_channels=inter_channels,
                                              dimension=3, sub_sample=sub_sample,
                                              bn_layer=bn_layer,)


if __name__ == '__main__':
    import torch

    for (sub_sample_, bn_layer_) in [(True, True), (False, False), (True, False), (False, True)]:
        img = torch.zeros(2, 3, 20)
        net = NONLocalBlock1D(3, sub_sample=sub_sample_, bn_layer=bn_layer_)
        out = net(img)
        print(out.size())

        img = torch.zeros(2, 3, 20, 20)
        net = NONLocalBlock2D(3, sub_sample=sub_sample_, bn_layer=bn_layer_)
        out = net(img)
        print(out.size())

        img = torch.randn(2, 3, 8, 20, 20)
        net = NONLocalBlock3D(3, sub_sample=sub_sample_, bn_layer=bn_layer_)
        out = net(img)
        print(out.size())





import os
import torch
from torch import nn
from tqdm import tqdm
from tensorboardX import SummaryWriter


class Solver(object):
    def __init__(self,
                 logs_path,
                 model: nn.Module,
                 train_loader,
                 valid_loader):
        self.model = model
        self.train_loader = train_loader
        self.valid_loader = valid_loader
        self.optim = torch.optim.Adam(self.model.parameters())
        self.criterion = nn.CrossEntropyLoss()

        self.iteration = 0
        tb_logs = os.path.join(logs_path, 'tb_logs')
        os.makedirs(tb_logs, exist_ok=True)
        self.writer = SummaryWriter(logdir=tb_logs)

        ckpt_path = os.path.join(logs_path, 'ckpt')
        os.makedirs(ckpt_path, exist_ok=True)
        self.ckpt_path = ckpt_path

    def save(self):
        torch.save(self.model.state_dict(), os.path.join(self.ckpt_path, f'model-{self.iteration}.pt'))

    @torch.no_grad()
    def test(self):
        mloss = 0
        mi = 0
        acc = 0
        for x, labels in tqdm(self.valid_loader):
            bs = len(x[1])
            logits = self.model(x)
            acc += torch.sum(torch.eq(torch.argmax(logits, dim=-1), labels).to(dtype=torch.float64)).item()
            loss = self.criterion(logits, labels)
            mloss += loss.item() * bs
            mi += bs
        mloss = mloss / (mi + (mi == 0))
        acc = acc / (mi + (mi == 0))
        self.writer.add_scalar('test/loss', mloss, self.iteration)
        self.writer.add_scalar('test/acc', acc, self.iteration)

    def train(self):
        for x, labels in tqdm(self.train_loader):
            self.optim.zero_grad()
            logits = self.model(x)
            loss = self.criterion(logits, labels)
            self.writer.add_scalar('train/loss', loss.item(), self.iteration)
            self.writer.add_scalar('train/acc', torch.eq(torch.argmax(logits, dim=-1), labels).mean(), self.iteration)
            loss.backward()
            self.optim.step()
            self.iteration += 1



import os
import sys
import json
import pickle
import torch
import shutil
import glob
import numpy as np


def remove_module_from_checkpoint_state_dict(state_dict):
    """
    Removes the prefix `module` from weight names that gets added by
    torch.nn.DataParallel()
    """
    from collections import OrderedDict
    new_state_dict = OrderedDict()
    for k, v in state_dict.items():
        name = k[7:]  # remove `module.`
        new_state_dict[name] = v
    return new_state_dict


def load_json_config(path):
    """ loads a json config file"""
    with open(path) as data_file:
        config = json.load(data_file)
        config = config_init(config)
    return config


def config_init(config):
    """ Some of the variables that should exist and contain default values """
    if "augmentation_mappings_json" not in config:
        config["augmentation_mappings_json"] = None
    if "augmentation_types_todo" not in config:
        config["augmentation_types_todo"] = None
    return config


def setup_cuda_devices(args):
    device_ids = []
    device = torch.device("cuda" if args.use_cuda else "cpu")
    if device.type == "cuda":
        device_ids = [int(i) for i in args.gpus.split(',')]
    return device, device_ids


def save_checkpoint(state, is_best, config, filename='checkpoint.pth.tar'):
    checkpoint_path = os.path.join(config['output_dir'], config['model_name'], filename)
    model_path = os.path.join(config['output_dir'], config['model_name'], 'model_best.pth.tar')
    torch.save(state, checkpoint_path)
    if is_best:
        print(" > Best model found at this epoch. Saving ...")
        shutil.copyfile(checkpoint_path, model_path)


def save_results(logits_matrix, features_matrix, targets_list, item_id_list,
                 class_to_idx, config):
    """
    Saves the predicted logits matrix, true labels, sample ids and class
    dictionary for further analysis of results
    """
    print("Saving inference results ...")
    path_to_save = os.path.join(
        config['output_dir'], config['model_name'], "test_results.pkl")
    with open(path_to_save, "wb") as f:
        pickle.dump([logits_matrix, features_matrix, targets_list,
                     item_id_list, class_to_idx], f)


def save_images_for_debug(dir_img, imgs):
    """
    2x3x12x224x224 --> [BS, C, seq_len, H, W]
    """
    print("Saving images to {}".format(dir_img))
    from matplotlib import pylab as plt
    imgs = imgs.permute(0, 2, 3, 4, 1)  # [BS, seq_len, H, W, C]
    imgs = imgs.mul(255).numpy()
    if not os.path.exists(dir_img):
        os.makedirs(dir_img)
    print(imgs.shape)
    for batch_id, batch in enumerate(imgs):
        batch_dir = os.path.join(dir_img, "batch{}".format(batch_id + 1))
        if not os.path.exists(batch_dir):
            os.makedirs(batch_dir)
        for j, img in enumerate(batch):
            plt.imsave(os.path.join(batch_dir, "frame{%04d}.png" % (j + 1)),
                       img.astype("uint8"))


def get_submission(logits_matrix, item_id_list, class_to_idx, config):
    top5_classes_pred_list = []

    for i, id in enumerate(item_id_list):
        logits_sample = logits_matrix[i]
        logits_sample_top5  = logits_sample.argsort()[-5:][::-1]
        # top1_class_index = logits_sample.argmax()
        # top1_class_label = class_to_idx[top1_class_index]

        top5_classes_pred_list.append(logits_sample_top5)

    path_to_save = os.path.join(
            config['output_dir'], config['model_name'], "test_submission.csv")
    with open(path_to_save, 'w') as fw:
        for id, top5_pred in zip(item_id_list, top5_classes_pred_list):
            fw.write("{}".format(id))
            for elem in top5_pred:
                fw.write(";{}".format(elem))
            fw.write("\n")


def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


class ExperimentalRunCleaner(object):
    """
    Remove the output dir, if you exit with Ctrl+C and if there are less
    then 1 file. It prevents the noise of experimental runs.
    """

    def __init__(self, save_dir):
        self.save_dir = save_dir

    def __call__(self, signal, frame):
        num_files = len(glob.glob(self.save_dir + "/*"))
        if num_files < 1:
            print('Removing: {}'.format(self.save_dir))
            shutil.rmtree(self.save_dir)
        print('You pressed Ctrl+C!')
        sys.exit(0)



import os
import json

from collections import namedtuple

ListData = namedtuple('ListData', ['id', 'label', 'path'])


class DatasetBase(object):
    """
    To read json data and construct a list containing video sample `ids`,
    `label` and `path`
    """
    def __init__(self, json_path_input, json_path_labels, data_root,
                 extension, is_test=False):
        self.json_path_input = json_path_input
        self.json_path_labels = json_path_labels
        self.data_root = data_root
        self.extension = extension
        self.is_test = is_test

        # preparing data and class dictionary
        self.classes = self.read_json_labels()
        self.classes_dict = self.get_two_way_dict(self.classes)
        self.json_data = self.read_json_input()

    def read_json_input(self):
        json_data = []
        if not self.is_test:
            with open(self.json_path_input, 'rb') as jsonfile:
                json_reader = json.load(jsonfile)
                for elem in json_reader:
                    label = self.clean_template(elem['template'])
                    if label not in self.classes:
                        raise ValueError("Label mismatch! Please correct")
                    item = ListData(elem['id'],
                                    label,
                                    os.path.join(self.data_root,
                                                 elem['id'] + self.extension)
                                    )
                    json_data.append(item)
        else:
            with open(self.json_path_input, 'rb') as jsonfile:
                json_reader = json.load(jsonfile)
                for elem in json_reader:
                    # add a dummy label for all test samples
                    item = ListData(elem['id'],
                                    "Holding something",
                                    os.path.join(self.data_root,
                                                 elem['id'] + self.extension)
                                    )
                    json_data.append(item)
        return json_data

    def read_json_labels(self):
        classes = []
        with open(self.json_path_labels, 'rb') as jsonfile:
            json_reader = json.load(jsonfile)
            for elem in json_reader:
                classes.append(elem)
        return sorted(classes)

    def get_two_way_dict(self, classes):
        classes_dict = {}
        for i, item in enumerate(classes):
            classes_dict[item] = i
            classes_dict[i] = item
        return classes_dict

    def clean_template(self, template):
        """ Replaces instances of `[something]` --> `something`"""
        template = template.replace("[", "")
        template = template.replace("]", "")
        return template


class WebmDataset(DatasetBase):
    def __init__(self, json_path_input, json_path_labels, data_root,
                 is_test=False):
        EXTENSION = ".webm"
        super().__init__(json_path_input, json_path_labels, data_root,
                         EXTENSION, is_test)


class I3DFeatures(DatasetBase):
    def __init__(self, json_path_input, json_path_labels, data_root,
                 is_test=False):
        EXTENSION = ".npy"
        super().__init__(json_path_input, json_path_labels, data_root,
                         EXTENSION, is_test)


class ImageNetFeatures(DatasetBase):
    def __init__(self, json_path_input, json_path_labels, data_root,
                 is_test=False):
        EXTENSION = ".npy"
        super().__init__(json_path_input, json_path_labels, data_root,
                         EXTENSION, is_test)



import torch
import torch.nn as nn
from torch.nn.parameter import Parameter
from torch.nn import functional as F
import numpy as np


class ContextGatedConv2d(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True):
        super(ContextGatedConv2d, self).__init__(in_channels, out_channels, kernel_size, stride,
                                                 padding, dilation, groups, bias)
        if kernel_size == 1:
            self.ind = True
        else:
            self.ind = False
            self.oc = out_channels
            self.ks = kernel_size

            ws = kernel_size
            self.avg_pool = nn.AdaptiveAvgPool2d((ws, ws))
            self.num_lat = int((kernel_size * kernel_size) / 2 + 1)
            self.ce = nn.Linear(ws * ws, self.num_lat, False)
            self.ce_bn = nn.BatchNorm1d(in_channels)
            self.ci_bn2 = nn.BatchNorm1d(in_channels)
            self.act = nn.ReLU(inplace=True)
            if in_channels // 16:
                self.g = 16
            else:
                self.g = in_channels
            self.ci = nn.Linear(self.g, out_channels // (in_channels // self.g), bias=False)
            self.ci_bn = nn.BatchNorm1d(out_channels)
            self.gd = nn.Linear(self.num_lat, kernel_size * kernel_size, False)
            self.gd2 = nn.Linear(self.num_lat, kernel_size * kernel_size, False)
            self.unfold = nn.Unfold(kernel_size, dilation, padding, stride)
            self.sig = nn.Sigmoid()

    def forward(self, x):
        if self.ind:
            return F.conv2d(x, self.weight, self.bias, self.stride,
                            self.padding, self.dilation, self.groups)
        else:
            b, c, h, w = x.size()
            weight = self.weight
            gl = self.avg_pool(x).view(b, c, -1)
            out = self.ce(gl)
            ce2 = out
            out = self.ce_bn(out)
            out = self.act(out)
            out = self.gd(out)
            if self.g > 3:
                oc = self.ci(self.act(self.ci_bn2(ce2).view(b, c // self.g, self.g, -1).transpose(2, 3))).transpose(2, 3).contiguous()
            else:
                oc = self.ci(self.act(self.ci_bn2(ce2).transpose(2, 1))).transpose(2, 1).contiguous()
            oc = oc.view(b, self.oc, -1)
            oc = self.ci_bn(oc)
            oc = self.act(oc)
            oc = self.gd2(oc)
            out = self.sig(out.view(b, 1, c, self.ks, self.ks) + oc.view(b, self.oc, 1, self.ks, self.ks))
            x_un = self.unfold(x)
            b, _, l = x_un.size()
            out = (out * weight.unsqueeze(0)).view(b, self.oc, -1)
            return torch.matmul(out, x_un).view(b, self.oc, int(np.sqrt(l)), int(np.sqrt(l)))



from .model import TBlock


CFG1 = [
    TBlock('Block1',
           in_channels=3,
           out_channels=32,
           spatial_ksize=3,
           spatial_stride=1,
           temporal_ksize=2,
           temporal_stride=2,
           temporal_pool_type='max',
           cg_ksize=3,
           cg_stride=2),
    TBlock('Block1',
           in_channels=32,
           out_channels=64,
           spatial_ksize=3,
           spatial_stride=1,
           temporal_ksize=3,
           temporal_stride=1,
           temporal_pool_type='avg',
           cg_ksize=3,
           cg_stride=2),
    TBlock('Block2',
           in_channels=64,
           out_channels=128,
           temporal_ksize=3,
           temporal_stride=1),
    TBlock('Block3',
           in_channels=128,
           spatial_ksize=3,
           temporal_ksize=3),
    TBlock('Block3',
           in_channels=128,
           spatial_ksize=3,
           temporal_ksize=3),
    TBlock('Block3',
           in_channels=128,
           spatial_ksize=3,
           temporal_ksize=3),
    TBlock('Block4',
           in_channels=128,
           out_channels=256,
           temporal_ksize=3,
           temporal_stride=1,
           cg_ksize=3),
    TBlock('Block4',
           in_channels=256,
           out_channels=512,
           temporal_ksize=3,
           temporal_stride=1,
           cg_ksize=3),
]



import torch
import cv2
import numpy as np
import numbers
import collections
import random


class ComposeMix(object):
    r"""Composes several transforms together. It takes a list of
    transformations, where each element odf transform is a list with 2
    elements. First being the transform function itself, second being a string
    indicating whether it's an "img" or "vid" transform
    Args:
        transforms (List[Transform, "<type>"]): list of transforms to compose.
                                                <type> = "img" | "vid"
    Example:
        >>> transforms.ComposeMix([
        [RandomCropVideo(84), "vid"],
        [torchvision.transforms.ToTensor(), "img"],
        [torchvision.transforms.Normalize(
                   mean=[0.485, 0.456, 0.406],  # default values for imagenet
                   std=[0.229, 0.224, 0.225]), "img"]
    ])
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, imgs):
        for t in self.transforms:
            if t[1] == "img":
                for idx, img in enumerate(imgs):
                    imgs[idx] = t[0](img)
            elif t[1] == "vid":
                imgs = t[0](imgs)
            else:
                print("Please specify the transform type")
                raise ValueError
        return imgs


class RandomCropVideo(object):
    r"""Crop the given video frames at a random location. Crop location is the
    same for all the frames.
    Args:
        size (sequence or int): Desired output size of the crop. If size is an
            int instead of sequence like (w, h), a square crop (size, size) is
            made.
        padding (int or sequence, optional): Optional padding on each border
            of the image. Default is 0, i.e no padding. If a sequence of length
            4 is provided, it is used to pad left, top, right, bottom borders
            respectively.
        pad_method (cv2 constant): Method to be used for padding.
    """

    def __init__(self, size, padding=0, pad_method=cv2.BORDER_CONSTANT):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size
        self.padding = padding
        self.pad_method = pad_method

    def __call__(self, imgs):
        """
        Args:
            img (numpy.array): Video to be cropped.
        Returns:
            numpy.array: Cropped video.
        """
        th, tw = self.size
        h, w = imgs[0].shape[:2]
        x1 = np.random.randint(0, w - tw)
        y1 = np.random.randint(0, h - th)
        for idx, img in enumerate(imgs):
            if self.padding > 0:
                img = cv2.copyMakeBorder(img, self.padding, self.padding,
                                         self.padding, self.padding,
                                         self.pad_method)
            # sample crop locations if not given
            # it is necessary to keep cropping same in a video
            img_crop = img[y1:y1 + th, x1:x1 + tw]
            imgs[idx] = img_crop
        return imgs


class RandomHorizontalFlipVideo(object):
    """Horizontally flip the given video frames randomly with a given probability.
    Args:
        p (float): probability of the image being flipped. Default value is 0.5
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, imgs):
        """
        Args:
            imgs (numpy.array): Video to be flipped.
        Returns:
            numpy.array: Randomly flipped video.
        """
        if random.random() < self.p:
            for idx, img in enumerate(imgs):
                imgs[idx] = cv2.flip(img, 1)
        return imgs

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


class RandomReverseTimeVideo(object):
    """Reverse the given video frames in time randomly with a given probability.
    Args:
        p (float): probability of the image being flipped. Default value is 0.5
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, imgs):
        """
        Args:
            imgs (numpy.array): Video to be flipped.
        Returns:
            numpy.array: Randomly flipped video.
        """
        if random.random() < self.p:
            imgs = imgs[::-1]
        return imgs

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


class RandomRotationVideo(object):
    """Rotate the given video frames randomly with a given degree.
    Args:
        degree (float): degrees used to rotate the video
    """

    def __init__(self, degree=10):
        self.degree = degree

    def __call__(self, imgs):
        """
        Args:
            imgs (numpy.array): Video to be rotated.
        Returns:
            numpy.array: Randomly rotated video.
        """
        h, w = imgs[0].shape[:2]
        degree_sampled = np.random.choice(
                            np.arange(-self.degree, self.degree, 0.5))
        M = cv2.getRotationMatrix2D((w / 2, h / 2), degree_sampled, 1)

        for idx, img in enumerate(imgs):
            imgs[idx] = cv2.warpAffine(img, M, (w, h))

        return imgs

    def __repr__(self):
        return self.__class__.__name__ + '(degree={})'.format(self.degree_sampled)


class IdentityTransform(object):
    """
    Returns same video back
    """
    def __init__(self,):
        pass

    def __call__(self, imgs):
        return imgs


class Scale(object):
    r"""Rescale the input image to the given size.
    Args:
        size (sequence or int): Desired output size. If size is a sequence like
            (w, h), output size will be matched to this. If size is an int,
            smaller edge of the image will be matched to this number.
            i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        interpolation (int, optional): Desired interpolation. Default is
            ``cv2.INTER_LINEAR``
    """

    def __init__(self, size, interpolation=cv2.INTER_LINEAR):
        assert isinstance(size, int) or (isinstance(
            size, collections.Iterable) and len(size) == 2)
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img):
        """
        Args:
            img (numpy.array): Image to be scaled.
        Returns:
            numpy.array: Rescaled image.
        """
        if isinstance(self.size, int):
            h, w = img.shape[:2]
            if (w <= h and w == self.size) or (h <= w and h == self.size):
                return img
            if w < h:
                ow = self.size
                oh = int(self.size * h / w)
                if ow < w:
                    return cv2.resize(img, (ow, oh), cv2.INTER_AREA)
                else:
                    return cv2.resize(img, (ow, oh))
            else:
                oh = self.size
                ow = int(self.size * w / h)
                if oh < h:
                    return cv2.resize(img, (ow, oh), cv2.INTER_AREA)
                else:
                    return cv2.resize(img, (ow, oh))
        else:
            return cv2.resize(img, tuple(self.size))


class UnNormalize(object):
    """Unnormalize an tensor image with mean and standard deviation.
    Given mean: (R, G, B) and std: (R, G, B),
    will normalize each channel of the torch.*Tensor, i.e.
    channel = (channel x std) + mean
    Args:
        mean (sequence): Sequence of means for R, G, B channels respecitvely.
        std (sequence): Sequence of standard deviations for R, G, B channels
            respecitvely.
    """

    def __init__(self, mean, std):
        self.mean = np.array(mean).astype('float32')
        self.std = np.array(std).astype('float32')

    def __call__(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
        Returns:
            Tensor: Normalized image.
        """
        if isinstance(tensor, torch.Tensor):
            self.mean = torch.FloatTensor(self.mean)
            self.std = torch.FloatTensor(self.std)

            if (self.std.dim() != tensor.dim() or
                    self.mean.dim() != tensor.dim()):
                for i in range(tensor.dim() - self.std.dim()):
                    self.std = self.std.unsqueeze(-1)
                    self.mean = self.mean.unsqueeze(-1)

            tensor = torch.add(torch.mul(tensor, self.std), self.mean)
        else:
            # Relying on Numpy broadcasting abilities
            tensor = tensor * self.std + self.mean
        return tensor



import torch.nn as nn
from torch.hub import load_state_dict_from_url
from torchvision.models import ResNet
from .se_layer import SELayer


def conv3x3(in_planes, out_planes, stride=1):
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)


class SEBasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None, reduction=16):
        super(SEBasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes, 1)
        self.bn2 = nn.BatchNorm2d(planes)
        self.se = SELayer(planes, reduction)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.se(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class SEBottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None, reduction=16):
        super(SEBottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.se = SELayer(planes * 4, reduction)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)
        out = self.se(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out



from torch import nn
# from .non_local_concatenation import NONLocalBlock2D
# from .non_local_gaussian import NONLocalBlock2D
from .non_local_embedded_gaussian import NONLocalBlock2D
# from .non_local_dot_product import NONLocalBlock2D


class Network(nn.Module):
    def __init__(self):
        super(Network, self).__init__()

        self.conv_1 = nn.Sequential(
            nn.Conv2d(in_channels=1, out_channels=32, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(32),
            nn.ReLU(),
            nn.MaxPool2d(2),
        )

        self.nl_1 = NONLocalBlock2D(in_channels=32)
        self.conv_2 = nn.Sequential(
            nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.MaxPool2d(2),
        )

        self.nl_2 = NONLocalBlock2D(in_channels=64)
        self.conv_3 = nn.Sequential(
            nn.Conv2d(in_channels=64, out_channels=128, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(128),
            nn.ReLU(),
            nn.MaxPool2d(2),
        )

        self.fc = nn.Sequential(
            nn.Linear(in_features=128*3*3, out_features=256),
            nn.ReLU(),
            nn.Dropout(0.5),

            nn.Linear(in_features=256, out_features=10)
        )

    def forward(self, x):
        batch_size = x.size(0)

        feature_1 = self.conv_1(x)
        nl_feature_1 = self.nl_1(feature_1)

        feature_2 = self.conv_2(nl_feature_1)
        nl_feature_2 = self.nl_2(feature_2)

        output = self.conv_3(nl_feature_2).view(batch_size, -1)
        output = self.fc(output)

        return output

    def forward_with_nl_map(self, x):
        batch_size = x.size(0)

        feature_1 = self.conv_1(x)
        nl_feature_1, nl_map_1 = self.nl_1(feature_1, return_nl_map=True)

        feature_2 = self.conv_2(nl_feature_1)
        nl_feature_2, nl_map_2 = self.nl_2(feature_2, return_nl_map=True)

        output = self.conv_3(nl_feature_2).view(batch_size, -1)
        output = self.fc(output)

        return output, [nl_map_1, nl_map_2]


if __name__ == '__main__':
    import torch

    img = torch.randn(3, 1, 28, 28)
    net = Network()
    out = net(img)
    print(out.size())



import sh
import re
import numpy as np
import torch

from .data_parser import WebmDataset
from .data_augmentor import Augmentor
import torchvision
from .transforms_video import *
from .utils import save_images_for_debug

from skvideo.io import FFmpegReader

ffprobe = sh.ffprobe.bake('-v', 'error', '-show_entries',
                          'format=start_time,duration')

FRAMERATE = 12  # default value


class VideoFolder(torch.utils.data.Dataset):

    def __init__(self, root, json_file_input, json_file_labels, clip_size,
                 nclips, step_size, is_val, transform_pre=None, transform_post=None,
                 augmentation_mappings_json=None, augmentation_types_todo=None,
                 get_item_id=False, is_test=False):
        self.dataset_object = WebmDataset(json_file_input, json_file_labels,
                                          root, is_test=is_test)
        self.json_data = self.dataset_object.json_data
        self.classes = self.dataset_object.classes
        self.classes_dict = self.dataset_object.classes_dict
        self.root = root
        self.transform_pre = transform_pre
        self.transform_post = transform_post

        self.augmentor = Augmentor(augmentation_mappings_json,
                                   augmentation_types_todo)

        self.clip_size = clip_size
        self.nclips = nclips
        self.step_size = step_size
        self.is_val = is_val
        self.get_item_id = get_item_id

    def __getitem__(self, index):
        item = self.json_data[index]

        framerate_sampled = self.augmentor.jitter_fps(FRAMERATE)

        optional_args = {"-r": "%d" % framerate_sampled}
        duration = self.get_duration(item.path)

        if duration is not None:
            nframes = int(duration * framerate_sampled)
            optional_args["-vframes"] = "%d" % nframes

        # Open video file
        reader = FFmpegReader(item.path, inputdict={}, outputdict=optional_args)

        try:
            imgs = []
            for img in reader.nextFrame():
                imgs.append(img)
        except (RuntimeError, ZeroDivisionError) as exception:
            print('{}: WEBM reader cannot open {}. Empty '
                  'list returned.'.format(type(exception).__name__, item.path))

        imgs = self.transform_pre(imgs)
        imgs, label = self.augmentor(imgs, item.label)
        imgs = self.transform_post(imgs)

        num_frames = len(imgs)
        target_idx = self.classes_dict[label]

        if self.nclips > -1:
            num_frames_necessary = self.clip_size * self.nclips * self.step_size
        else:
            num_frames_necessary = num_frames
        offset = 0
        if num_frames_necessary < num_frames:
            # If there are more frames, then sample starting offset.
            diff = (num_frames - num_frames_necessary)
            # temporal augmentation
            if not self.is_val:
                offset = np.random.randint(0, diff)

        imgs = imgs[offset: num_frames_necessary + offset: self.step_size]

        if len(imgs) < (self.clip_size * self.nclips):
            imgs.extend([imgs[-1]] *
                        ((self.clip_size * self.nclips) - len(imgs)))

        # format data to torch
        data = torch.stack(imgs)
        data = data.permute(1, 0, 2, 3)
        if self.get_item_id:
            return (data, target_idx, item.id)
        else:
            return (data, target_idx)

    def __len__(self):
        return len(self.json_data)

    def get_duration(self, file):
        cmd_output = ffprobe(file)
        start_time, duration = re.findall("\d+\.\d+", str(cmd_output.stdout))
        return float(duration) - float(start_time)


if __name__ == '__main__':
    upscale_size = int(84 * 1.1)
    transform_pre = ComposeMix([
            # [RandomRotationVideo(20), "vid"],
            [Scale(upscale_size), "img"],
            [RandomCropVideo(84), "vid"],
            # [RandomHorizontalFlipVideo(0), "vid"],
            # [RandomReverseTimeVideo(1), "vid"],
            # [torchvision.transforms.ToTensor(), "img"],
             ])
    # identity transform
    transform_post = ComposeMix([
                        [torchvision.transforms.ToTensor(), "img"],
                         ])

    loader = VideoFolder(root="/data-ssd1/20bn-something-something-v2/videos",
                         json_file_input="/data-ssd1/20bn-something-something-v2/annotations/something-something-v2-train.json",
                         json_file_labels="/data-ssd1/20bn-something-something-v2/annotations/something-something-v2-labels.json",
                         clip_size=36,
                         nclips=1,
                         step_size=1,
                         is_val=False,
                         transform_pre=transform_pre,
                         transform_post=transform_post,
                         # augmentation_mappings_json="notebooks/augmentation_mappings.json",
                         # augmentation_types_todo=["left/right", "left/right agnostic", "jitter_fps"],
                         )
    # fetch a sample
    # data_item, target_idx = loader[1]
    # save_images_for_debug("input_images_2", data_item.unsqueeze(0))
    # print("Label = {}".format(loader.classes_dict[target_idx]))

    import time
    from tqdm import tqdm

    batch_loader = torch.utils.data.DataLoader(
        loader,
        batch_size=10, shuffle=False,
        num_workers=8, pin_memory=True)

    start = time.time()
    for i, a in enumerate(tqdm(batch_loader)):
        if i > 100:
            break
        pass
    print("Size --> {}".format(a[0].size()))
    print(time.time() - start)



import os
import sys
from setuptools import setup


setup(name='tvn',
      version='1.0',
      description='Tiny Video Networks',
      author='George Kasparyants',
      author_email='gkasparyants@gmail.com',
      url='https://github.com/DELTA37/TVN.git',
      packages=['tvn',
                'tvn.cg',
                'tvn.non_local',
                'tvn.se',
                'tvn.temporal'],
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ])



import torch
from torch import nn
from torch.nn import functional as F


class Temporal(nn.Module):
    def __init__(self):
        super(Temporal, self).__init__()

    def forward_single(self, x: torch.Tensor):
        raise NotImplementedError()

    def forward(self, x, seq_lens):
        out = []
        new_seq_lens = []
        for i in range(len(seq_lens)):
            idx = sum(seq_lens[:i])
            length = seq_lens[i]

            if len(seq_lens) == 1:
                """
                Option for batch_size = 1
                """
                temp_x = x
            else:
                temp_x = x[idx: idx + length]

            y = self.forward_single(temp_x)

            new_seq_len = y.size(0)
            new_seq_lens.append(new_seq_len)

            out.append(y)

        if len(seq_lens) == 1:
            """
            Option for batch_size = 1
            """
            return out[0], new_seq_lens
        else:
            return torch.cat(out, dim=0), new_seq_lens


class TemporalMaxPool1d(Temporal):
    def __init__(self,
                 kernel_size,
                 stride=None,
                 padding=0):
        super(TemporalMaxPool1d, self).__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding

    def forward_single(self, x: torch.Tensor):
        seq_len, c, h, w = x.size()
        x = torch.transpose(x.view(seq_len, c, h * w), 2, 0)  # [bs, channels, L]
        y = F.max_pool1d(x,
                         kernel_size=self.kernel_size,
                         stride=self.stride,
                         padding=self.padding)
        y = torch.transpose(y, 2, 0).view(y.size(2), y.size(1), h, w)
        return y


class TemporalAvgPool1d(Temporal):
    def __init__(self,
                 kernel_size,
                 stride=None,
                 padding=0):
        super(TemporalAvgPool1d, self).__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding

    def forward_single(self, x: torch.Tensor):
        seq_len, c, h, w = x.size()
        x = torch.transpose(x.view(seq_len, c, h * w), 2, 0)  # [bs, channels, L]
        y = F.avg_pool1d(x,
                         kernel_size=self.kernel_size,
                         stride=self.stride,
                         padding=self.padding)
        y = torch.transpose(y, 2, 0).view(y.size(2), y.size(1), h, w)
        return y


class TemporalConv1d(Temporal):
    def __init__(self,
                 in_channels: int, out_channels: int, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1):
        super(TemporalConv1d, self).__init__()
        self.conv = nn.Conv1d(in_channels=in_channels,
                              out_channels=out_channels,
                              kernel_size=kernel_size,
                              stride=stride,
                              padding=padding,
                              dilation=dilation,
                              groups=groups)

    def forward_single(self, x: torch.Tensor):
        seq_len, c, h, w = x.size()
        y = torch.transpose(x.view(seq_len, c, h * w), 2, 0)  # [bs, channels, L]
        y = self.conv(y)
        y = torch.transpose(y, 2, 0).view(y.size(2), y.size(1), h, w)
        return y


class TemporalGlobalAvgPool(Temporal):
    def __init__(self, squeeze=True):
        super(TemporalGlobalAvgPool, self).__init__()
        self.squeeze = squeeze

    def forward_single(self, x: torch.Tensor):
        x = torch.mean(x, dim=[0, 2, 3], keepdim=True)
        if self.squeeze:
            x = x.squeeze(-1).squeeze(-1)
        return x



import json
import numpy as np

from .transforms_video import *


class Augmentor(object):
    def __init__(self, augmentation_mappings_json=None,
                 augmentation_types_todo=None,
                 fps_jitter_factors=[1, 0.75, 0.5]):
        self.augmentation_mappings_json = augmentation_mappings_json
        self.augmentation_types_todo = augmentation_types_todo
        self.fps_jitter_factors = fps_jitter_factors

        # read json to get the mapping dict
        self.augmentation_mapping = self.read_augmentation_mapping(
                                        self.augmentation_mappings_json)
        self.augmentation_transforms = self.define_augmentation_transforms()

    def __call__(self, imgs, label):
        if not self.augmentation_mapping:
            return imgs, label
        else:
            candidate_augmentations = {"same": label}
            for candidate in self.augmentation_types_todo:
                if candidate == "jitter_fps":
                    continue
                if label in self.augmentation_mapping[candidate]:
                    if isinstance(self.augmentation_mapping[candidate], list):
                        candidate_augmentations[candidate] = label
                    elif isinstance(self.augmentation_mapping[candidate], dict):
                        candidate_augmentations[candidate] = self.augmentation_mapping[candidate][label]
                    else:
                        print("Something wrong with data type specified in "
                              "augmentation file. Please check!")
            augmentation_chosen = np.random.choice(list(candidate_augmentations.keys()))
            imgs = self.augmentation_transforms[augmentation_chosen](imgs)
            label = candidate_augmentations[augmentation_chosen]

            return imgs, label

    def read_augmentation_mapping(self, path):
        if path:
            with open(path, "rb") as fp:
                mapping = json.load(fp)
        else:
            mapping = None
        return mapping

    def define_augmentation_transforms(self, ):
        augmentation_transforms = {}
        augmentation_transforms["same"] = IdentityTransform()
        augmentation_transforms["left/right"] = RandomHorizontalFlipVideo(1)
        augmentation_transforms["left/right agnostic"] = RandomHorizontalFlipVideo(1)
        augmentation_transforms["reverse time"] = RandomReverseTimeVideo(1)
        augmentation_transforms["reverse time agnostic"] = RandomReverseTimeVideo(0.5)

        return augmentation_transforms

    def jitter_fps(self, framerate):
        if self.augmentation_types_todo and "jitter_fps" in self.augmentation_types_todo:
            jitter_factor = np.random.choice(self.fps_jitter_factors)
            return int(jitter_factor * framerate)
        else:
            return framerate



import torch
from torch import nn
from torch.nn import functional as F
from .temporal.temporal_blocks import TemporalAvgPool1d, TemporalMaxPool1d, TemporalConv1d, TemporalGlobalAvgPool
from .cg.cg_block import ContextGatedConv2d
from .se.se_block import SEBasicBlock
from .non_local.non_local_embedded_gaussian import NONLocalBlock2D
from collections import namedtuple


TBlock = namedtuple('TBlock', [
    'name',
    'in_channels',
    'out_channels',
    'spatial_ksize',
    'spatial_stride',
    'temporal_ksize',
    'temporal_stride',
    'temporal_pool_type',
    'cg_ksize',
    'cg_stride',
])
TBlock.__new__.__defaults__ = (None,) * len(TBlock._fields)
VERBOSE = True


def no_verbose():
    global VERBOSE
    VERBOSE = False


def verbose():
    global VERBOSE
    VERBOSE = True


def create_block(t_block: TBlock) -> nn.Module:
    if t_block.name == 'Block1':
        return Block1(in_channels=t_block.in_channels,
                      out_channels=t_block.out_channels,
                      spatial_ksize=t_block.spatial_ksize,
                      spatial_stride=t_block.spatial_stride,
                      temporal_ksize=t_block.temporal_ksize,
                      temporal_stride=t_block.temporal_stride,
                      temporal_pool_type=t_block.temporal_pool_type,
                      cg_ksize=t_block.cg_ksize,
                      cg_stride=t_block.cg_stride)
    elif t_block.name == 'Block2':
        return Block2(in_channels=t_block.in_channels,
                      out_channels=t_block.out_channels,
                      temporal_ksize=t_block.temporal_ksize,
                      temporal_stride=t_block.temporal_stride)
    elif t_block.name == 'Block3':
        return Block3(in_channels=t_block.in_channels,
                      spatial_ksize=t_block.spatial_ksize,
                      temporal_ksize=t_block.temporal_ksize)
    elif t_block.name == 'Block4':
        return Block4(in_channels=t_block.in_channels,
                      out_channels=t_block.out_channels,
                      temporal_ksize=t_block.temporal_ksize,
                      temporal_stride=t_block.temporal_stride,
                      cg_ksize=t_block.cg_ksize)
    else:
        raise NotImplementedError()


class Block1(nn.Module):
    def __init__(self, in_channels, out_channels,
                 spatial_ksize=3,
                 spatial_stride=1,
                 temporal_ksize=2,
                 temporal_stride=2,
                 temporal_pool_type='max',
                 cg_ksize=3,
                 cg_stride=2):
        super(Block1, self).__init__()
        self.spatial_conv = nn.Conv2d(in_channels, out_channels,
                                      kernel_size=spatial_ksize,
                                      stride=spatial_stride,
                                      padding=(spatial_ksize - 1) // 2)

        if temporal_pool_type == 'max':
            self.pool1d = TemporalMaxPool1d(temporal_ksize,
                                            stride=temporal_stride,
                                            padding=(temporal_ksize - 1) // 2)
        elif temporal_pool_type == 'avg':
            self.pool1d = TemporalAvgPool1d(temporal_ksize,
                                            stride=temporal_stride,
                                            padding=(temporal_ksize - 1) // 2)
        else:
            raise NotImplementedError()

        self.cg = ContextGatedConv2d(out_channels, out_channels,
                                     kernel_size=cg_ksize,
                                     stride=cg_stride,
                                     padding=(cg_ksize - 1) // 2)

    def forward(self, input):
        x, seq_lens = input
        x = self.spatial_conv(x)
        if VERBOSE:
            print('Block1 inputs', x.shape, seq_lens)
        x, seq_lens = self.pool1d(x, seq_lens)
        if VERBOSE:
            print('Block1 outputs', x.shape, seq_lens)
        x = self.cg(x)
        return x, seq_lens


class Block2(nn.Module):
    def __init__(self, in_channels, out_channels,
                 temporal_ksize=3,
                 temporal_stride=2):
        super(Block2, self).__init__()

        self.conv1d = TemporalConv1d(in_channels, out_channels,
                                     kernel_size=temporal_ksize,
                                     stride=temporal_stride,
                                     padding=(temporal_ksize - 1) // 2)
        self.se = SEBasicBlock(out_channels, out_channels)

    def forward(self, input):
        x, seq_lens = input
        x = F.adaptive_avg_pool2d(x, (x.size(2) // 2, x.size(3) // 2))
        if VERBOSE:
            print('Block2 inputs', x.shape, seq_lens)
        x, seq_lens = self.conv1d(x, seq_lens)
        if VERBOSE:
            print('Block2 outputs', x.shape, seq_lens)
        x = self.se(x)
        return x, seq_lens


class Block3(nn.Module):
    def __init__(self, in_channels,
                 spatial_ksize=3,
                 temporal_ksize=3):
        super(Block3, self).__init__()

        self.spatial_conv2d = nn.Conv2d(in_channels, in_channels,
                                        kernel_size=spatial_ksize,
                                        stride=1,
                                        padding=(spatial_ksize - 1) // 2)

        self.conv1d = TemporalConv1d(in_channels, in_channels,
                                     kernel_size=temporal_ksize,
                                     stride=1,
                                     padding=(temporal_ksize - 1) // 2)

        self.nl = NONLocalBlock2D(in_channels)

    def forward(self, input):
        x, seq_lens = input
        residual_x = x
        x = self.spatial_conv2d(x)
        if VERBOSE:
            print('Block3 inputs', x.shape, seq_lens)
        x, seq_lens = self.conv1d(x, seq_lens)
        if VERBOSE:
            print('Block3 outputs', x.shape, seq_lens)
        x = self.nl(x)
        x = residual_x + x
        return x, seq_lens


class Block4(nn.Module):
    def __init__(self, in_channels, out_channels,
                 temporal_ksize=3,
                 temporal_stride=2,
                 cg_ksize=3):
        super(Block4, self).__init__()

        self.conv1d = TemporalConv1d(in_channels, out_channels,
                                     kernel_size=temporal_ksize,
                                     stride=temporal_stride,
                                     padding=(temporal_ksize - 1) // 3)

        self.cg = ContextGatedConv2d(out_channels, out_channels,
                                     kernel_size=cg_ksize,
                                     stride=1,
                                     padding=(cg_ksize - 1) // 2)

        self.se = SEBasicBlock(out_channels, out_channels)

    def forward(self, input):
        x, seq_lens = input
        x = F.adaptive_avg_pool2d(x, (x.size(2) // 2, x.size(3) // 2))
        if VERBOSE:
            print('Block4 inputs', x.shape, seq_lens)
        x, seq_lens = self.conv1d(x, seq_lens)
        if VERBOSE:
            print('Block4 outputs', x.shape, seq_lens)
        x = self.cg(x)
        x = self.se(x)
        return x, seq_lens


class TVN(nn.Module):
    def __init__(self, t_blocks,
                 num_classes,
                 prepare_seq=True):
        super(TVN, self).__init__()
        self.body = nn.Sequential(*[create_block(t_block) for t_block in t_blocks])
        self.global_pool = TemporalGlobalAvgPool(squeeze=True)
        self.head = nn.Sequential(nn.Linear(512, 512),
                                  nn.LeakyReLU(0.2),
                                  nn.Linear(512, num_classes))
        self.prepare_seq = prepare_seq

    def forward(self, input,
                extract_feas=False):
        """

        :param input: (x, seq_lens): [sum(seq_lens), H, W, C], idx: idx + seq_len - frames from unique video
        :param extract_feas
        :return:
        """
        if self.prepare_seq:
            bs, c, l, h, w = input.size()
            input = (input.permute(0, 2, 1, 3, 4).reshape(bs * l, c, h, w), [l for _ in range(bs)])

        x, seq_lens = input
        x, seq_lens = self.body((x, seq_lens))
        feas, seq_lens = self.global_pool(x, seq_lens)
        if VERBOSE:
            print(seq_lens)
        assert all([seq_len == 1 for seq_len in seq_lens])
        if extract_feas:
            return feas
        logits = self.head(feas)
        return logits
